@@ -25,7 +25,7 @@ class RandomArrayTest:
 
 
 class RandomArrayTestObject:
-    def __init__(self, input_arr, kernel, output_arr, padding, stride=1):
+    def __init__(self, input_arr, kernel, output_arr, bias, padding, stride=1):
         """Struct that represents one test case for the random array tests.
         input_arr: ndarray, 3-Dimensional floating point numpy array
         output_arr: ndarray, 3-Dimensional floating point numpy array
@@ -45,15 +45,20 @@ class RandomArrayTestObject:
         self.input_arr = numpy_array_to_rust(input_arr, shape_vec=True)
         self.output_arr = numpy_array_to_rust(output_arr, shape_vec=True)
         self.kernel = numpy_array_to_rust(kernel, shape_vec=True)
+        if bias is None:
+            self.bias = "None"
+        else:
+            self.bias = numpy_array_to_rust(
+                bias, shape_vec=True, bias=True)
         self.stride = stride
 
 
-def numpy_array_to_rust(x, shape_vec=False):
+def numpy_array_to_rust(x, shape_vec=False, bias=False):
     """
         Outputs a numpy array as a Rust ndarray.
-        If shape_vec is set to true, outputs 
+        If shape_vec is set to true, outputs
         the array creationg through the shape_vec Rust function.
-        The Rust array macro seems broken for 4-D arrays, so this is a 
+        The Rust array macro seems broken for 4-D arrays, so this is a
         workaround.
     """
     # This removes the "dtype=..." info in the representation,
@@ -71,7 +76,10 @@ def numpy_array_to_rust(x, shape_vec=False):
     # removes leading array and closing paren tokens
     array_repr = f"{repr(x)}"[6:][:ending_delimiter].replace("\n", "\n\t\t")
     if shape_vec:
-        return f"Array::from_shape_vec({x_shape}, vec!{array_repr}).unwrap()"
+        if bias:
+            return f"Some(Array::from_shape_vec({x_shape}, vec!{array_repr}).unwrap())"
+        else:
+            return f"Array::from_shape_vec({x_shape}, vec!{array_repr}).unwrap()"
     else:
         return f"array!{array_repr}".rstrip().rstrip(",")
 
@@ -128,7 +136,7 @@ def transform(orig, dest, x, is_kernel):
     orig: str, tf, rust, or pt
     dest: str, tf, rust, or pt
     x: ndarray to transform
-    transpose: whether 
+    transpose: whether
     kernel: bool, True if x is a kernel
     """
     orig = orig.lower()
@@ -165,8 +173,8 @@ def transform(orig, dest, x, is_kernel):
     return x
 
 
-def conv2d_random_array_test(img_shapes, kernel_shapes, num_arrays_per_case=3, use_torch=False, transpose=False, seed=260896, padding="VALID", stride=1, compare_impls=True):
-    """Returns a Test case that can be rendered with the 
+def conv2d_random_array_test(img_shapes, kernel_shapes, num_arrays_per_case=3, use_torch=False, transpose=False, seed=260896, padding="VALID", stride=1, bias=False, compare_impls=True):
+    """Returns a Test case that can be rendered with the
     test_py_impl_random_arrays_template.rs into a Rust test
     that tests the conv2d Rust implementation against tf.nn.conv2d.
 
@@ -208,28 +216,40 @@ def conv2d_random_array_test(img_shapes, kernel_shapes, num_arrays_per_case=3, u
         for i in range(num_arrays_per_case):
             im = np.random.rand(*im_shape).astype(dtype=np.float32)
             ker = np.random.rand(*ker_shape).astype(dtype=np.float32)
+            if bias:
+                if transpose:
+                    bias_shape = ker_shape[1]
+                else:
+                    bias_shape = ker_shape[0]
+                bias_vec = np.random.rand(bias_shape).astype(dtype=np.float32)
+            else:
+                bias_vec = None
 
             # Generating the PT and TF parts
 
             if compare_impls or use_torch:
                 im_pt = torch.FloatTensor(transform_img("rust", "pt", im))
                 ker_pt = torch.FloatTensor(transform_ker("rust", "pt", ker))
+                if bias:
+                    bias_pt = torch.FloatTensor(bias_vec)
+                else:
+                    bias_pt = None
 
                 if transpose:
                     if padding == "SAME":
                         # This simulates padding = "SAME"
                         out_pt = torch.nn.functional.conv_transpose2d(
-                            im_pt, ker_pt, output_padding=stride-1, padding=(ker_pt.shape[2] // 2, ker_pt.shape[3]//2), stride=stride)
+                            im_pt, ker_pt, bias_pt, output_padding=stride-1, padding=(ker_pt.shape[2] // 2, ker_pt.shape[3]//2), stride=stride)
                     else:
                         out_pt = torch.nn.functional.conv_transpose2d(
-                            im_pt, ker_pt, padding=0, stride=stride)
+                            im_pt, ker_pt, bias_pt, padding=0, stride=stride)
                 else:
                     if padding == "SAME" and stride > 1:
                         out_pt = torch.nn.functional.conv2d(
-                            im_pt, ker_pt, padding=(ker_pt.shape[2] // 2, ker_pt.shape[3]//2), stride=stride)
+                            im_pt, ker_pt, bias_pt, padding=(ker_pt.shape[2] // 2, ker_pt.shape[3]//2), stride=stride)
                     else:
                         out_pt = torch.nn.functional.conv2d(
-                            im_pt, ker_pt, padding=padding.lower(), stride=stride)
+                            im_pt, ker_pt, bias_pt, padding=padding.lower(), stride=stride)
 
                 out_pt_numpy = transform_img("pt", "rust", out_pt.numpy())
 
@@ -262,9 +282,11 @@ def conv2d_random_array_test(img_shapes, kernel_shapes, num_arrays_per_case=3, u
             else:
                 out = out_tf_numpy
 
+            # TODO() write out bias with shape (output_channel, 1)
+            # Rust expects Array2<F>
             # Writing the test objects
             test_obj = RandomArrayTestObject(
-                im, ker, out, padding, stride=stride)
+                im, ker, out, bias_vec, padding, stride=stride)
             objects.append(test_obj)
 
     if transpose:
@@ -309,6 +331,8 @@ def main():
     # Im shapes: Channels, Height, Width. Ker shapes are Out, In, Height, Width
     # (the format used in Rust)
 
+    # Note: Tensorflow does not provide an option to input bias; so bias is tested with pytorch
+
     img_shapes = [(1, 5, 12), (1, 10, 15), (1, 15, 10),
                   (3, 6, 6), (3, 10, 15), (3, 15, 10)]
     kernel_shapes = [(2, 1, 3, 4), (2, 1, 5, 5),
@@ -331,20 +355,13 @@ def main():
     template = env.get_template("test_py_impl_random_arrays_template.rs")
     ml_test_folder = os.path.join(project_root, "tests")
 
-    # writing out the conv2d test cases
+    # stride = 1, padding= "VALID", normal conv
     conv2d_test_case = conv2d_random_array_test(
-        img_shapes, kernel_shapes)
+        img_shapes, kernel_shapes, padding="VALID", compare_impls=True)
     conv2d_test_content = template.render(
         random_tests=[conv2d_test_case], file=__file__)
-    write_test_to_file(ml_test_folder, conv2d_test_content, "conv2d")
-
-    # writing out the conv2d test cases with torch
-    conv2d_torch_test_case = conv2d_random_array_test(
-        img_shapes, kernel_shapes, use_torch=True)
-    conv2d_torch_test_content = template.render(
-        random_tests=[conv2d_torch_test_case], file=__file__)
-    write_test_to_file(
-        ml_test_folder, conv2d_torch_test_content, "conv2d_torch")
+    write_test_to_file(ml_test_folder, conv2d_test_content,
+                       "conv2d_stride1_valid_both")
 
     # writing out the conv2d_stride2 test cases with stride 2
     # we need to skip the uneven kernel, as same padding in torch
@@ -358,39 +375,25 @@ def main():
     conv2d_stride2_test_content = template.render(
         random_tests=[conv2d_stride2_test_case], file=__file__)
     write_test_to_file(
-        ml_test_folder, conv2d_stride2_test_content, "conv2d_stride2_torch_same")
+        ml_test_folder, conv2d_stride2_test_content, "conv2d_stride2_same_torch")
 
-    conv2d_stride2_torch_test_case = conv2d_random_array_test(
-        img_shapes, kernel_shapes, use_torch=True, stride=2, padding="VALID")
-    conv2d_stride2_torch_test_content = template.render(
-        random_tests=[conv2d_stride2_torch_test_case], file=__file__)
-    write_test_to_file(
-        ml_test_folder, conv2d_stride2_torch_test_content, "conv2d_stride2_torch_valid")
+    # stride 2, padding = "VALID", conv2 transpose, no bias
+    conv2d_transpose_test_case = conv2d_random_array_test(
+        img_shapes_trans, kernel_shapes_trans, transpose=True, padding="VALID", compare_impls=False, use_torch=True, stride=2)
+    conv2d_transpose_test_content = template.render(
+        random_tests=[conv2d_transpose_test_case], file=__file__)
+    write_test_to_file(ml_test_folder, conv2d_transpose_test_content,
+                       "conv2d_transpose_stride2_valid_torch")
 
     # transpose, padding same, stride2. Uses torch, as impls cant be compared
     # only uneven kernel sizes usable
+    # stride 2, padding = "SAME", conv2 transpose, no bias
     conv2d_stride2_same_transpose_test_case = conv2d_random_array_test(
         img_shapes_trans, kernel_shapes_trans[2:], transpose=True, padding="SAME", compare_impls=False, use_torch=True, stride=2)
     conv2d_stride2_transpose_test_content = template.render(
         random_tests=[conv2d_stride2_same_transpose_test_case], file=__file__)
     write_test_to_file(ml_test_folder, conv2d_stride2_transpose_test_content,
-                       "conv2d_transpose_torch_stride2_same")
-
-    # transpose, padding valid, stride1
-    conv2d_transpose_test_case = conv2d_random_array_test(
-        img_shapes_trans, kernel_shapes_trans, transpose=True, padding="VALID", compare_impls=False, use_torch=True)
-    conv2d_transpose_test_content = template.render(
-        random_tests=[conv2d_transpose_test_case], file=__file__)
-    write_test_to_file(ml_test_folder, conv2d_transpose_test_content,
-                       "conv2d_transpose_torch_valid")
-
-    # transpose, padding valid, stride2
-    conv2d_transpose_test_case = conv2d_random_array_test(
-        img_shapes_trans, kernel_shapes_trans, transpose=True, padding="VALID", compare_impls=False, stride=2, use_torch=True)
-    conv2d_transpose_test_content = template.render(
-        random_tests=[conv2d_transpose_test_case], file=__file__)
-    write_test_to_file(ml_test_folder, conv2d_transpose_test_content,
-                       "conv2d_transpose_torch_valid_stride2")
+                       "conv2d_transpose_stride2_same_torch")
 
     # writing out the conv2d_tranposed test cases for
     # a change in channel (output != input channel)
@@ -401,6 +404,22 @@ def main():
     write_test_to_file(ml_test_folder, conv2d_transpose_test_content,
                        "conv2d_transpose_stride2_test_different_in_out_channel")
 
+    # transpose, padding valid, stride1
+    conv2d_transpose_test_case = conv2d_random_array_test(
+        img_shapes_trans, kernel_shapes_trans, transpose=True, padding="VALID", compare_impls=False, use_torch=True, bias=True)
+    conv2d_transpose_test_content = template.render(
+        random_tests=[conv2d_transpose_test_case], file=__file__)
+    write_test_to_file(ml_test_folder, conv2d_transpose_test_content,
+                       "conv2d_transpose_stride1_valid_torch_bias")
+
+    # stride 1, padding = "VALID", normal conv, bias
+    conv2d_stride2_torch_test_case = conv2d_random_array_test(
+        img_shapes, kernel_shapes, use_torch=True, stride=1, padding="VALID", bias=True, compare_impls=False)
+    conv2d_stride2_torch_test_content = template.render(
+        random_tests=[conv2d_stride2_torch_test_case], file=__file__)
+    write_test_to_file(
+        ml_test_folder, conv2d_stride2_torch_test_content, "conv2d_stride2_valid_torch_bias")
+
 
 if __name__ == "__main__":
     main()
