@@ -13,19 +13,30 @@ use num_traits::Float;
 pub struct ConvolutionLayer<F: Float> {
     /// Weight matrix of the kernel
     pub(in crate) kernel: Array4<F>,
+    pub(in crate) bias: Option<Array1<F>>,
     pub(in crate) stride: usize,
     pub(in crate) padding: Padding,
 }
 
-impl<F: 'static + Float> ConvolutionLayer<F> {
-    /// Creates new convolution layer. The weights are given in
-    /// Pytorch layout.
+impl<F: 'static + Float + std::ops::AddAssign> ConvolutionLayer<F> {
+    /// Creates new convolution layer.
+    /// The weights are given in Pytorch layout.
     /// (out channels, in channels, kernel height, kernel width)
-    pub fn new(weights: Array4<F>, stride: usize, padding: Padding) -> ConvolutionLayer<F> {
+    /// Bias: (output height * output width, 1)
+    pub fn new(
+        weights: Array4<F>,
+        bias_array: Option<Array1<F>>,
+        stride: usize,
+        padding: Padding,
+    ) -> ConvolutionLayer<F> {
         assert!(stride > 0, "Stride of 0 passed");
-
+        // match bias_array{
+        //     Some(x) => x,
+        //     None => None,
+        // }
         ConvolutionLayer {
             kernel: weights,
+            bias: bias_array,
             stride,
             padding,
         }
@@ -34,19 +45,30 @@ impl<F: 'static + Float> ConvolutionLayer<F> {
     /// Creates new convolution layer. The weights are given in
     /// Tensorflow layout.
     /// (kernel height, kernel width, in channels, out channels)
-    pub fn new_tf(weights: Array4<F>, stride: usize, padding: Padding) -> ConvolutionLayer<F> {
+    pub fn new_tf(
+        weights: Array4<F>,
+        bias_array: Option<Array1<F>>,
+        stride: usize,
+        padding: Padding,
+    ) -> ConvolutionLayer<F> {
         let permuted_view = weights.view().permuted_axes([3, 2, 0, 1]);
         // Hack to fix the memory layout, permuted axes makes a
         // col major array / non-contiguous array from weights
         let permuted_array: Array4<F> =
             Array::from_shape_vec(permuted_view.dim(), permuted_view.iter().copied().collect())
                 .unwrap();
-        ConvolutionLayer::new(permuted_array, stride, padding)
+        ConvolutionLayer::new(permuted_array, bias_array, stride, padding)
     }
 
     /// Analog to conv2d.
     pub fn convolve(&self, image: &DataRepresentation<F>) -> DataRepresentation<F> {
-        conv2d(&self.kernel, image, self.padding, self.stride)
+        conv2d(
+            &self.kernel,
+            self.bias.as_ref(),
+            image,
+            self.padding,
+            self.stride,
+        )
     }
 }
 
@@ -175,8 +197,9 @@ where
 /// Returns:
 /// -----------------------------------------------
 /// - out: Output data, of shape (F, H', W')
-pub fn conv2d<'a, T, V, F: 'static + Float>(
+pub fn conv2d<'a, T, V, F: 'static + Float + std::ops::AddAssign>(
     kernel_weights: T,
+    bias: Option<&Array1<F>>,
     im2d: V,
     padding: Padding,
     stride: usize,
@@ -273,6 +296,35 @@ where
         );
     };
     let filter_transpose = filter_col.t();
-    let mul = im_col.dot(&filter_transpose); // + bias_m
-    col2im_ref(&mul, new_im_height, new_im_width, 1)
+    let mul = im_col.dot(&filter_transpose);
+    let output = col2im_ref(&mul, new_im_height, new_im_width, 1);
+    add_bias(&output, bias)
+}
+
+pub(in crate) fn add_bias<F>(x: &Array3<F>, bias: Option<&Array1<F>>) -> Array3<F>
+where
+    F: 'static + Float + std::ops::AddAssign,
+{
+    if let Some(bias_array) = bias {
+        assert!(
+            bias_array.shape()[0] == x.shape()[0],
+            "Bias array has the wrong shape {:?} for vec of shape {:?}",
+            bias_array.shape(),
+            x.shape()
+        );
+        // Yes this is really necessary. Broadcasting with ndarray-rust
+        // starts at the right side of the shape, so we have to add
+        // the axes by hand (else it thinks that it should compare the
+        // output width and the bias channels).
+        (x + &bias_array
+            .clone()
+            .insert_axis(Axis(1))
+            .insert_axis(Axis(2))
+            .broadcast(x.shape())
+            .unwrap())
+            .into_dimensionality()
+            .unwrap()
+    } else {
+        x.clone()
+    }
 }
