@@ -5,7 +5,7 @@
 //! - <https://github.com/akutzer/numpy_cnn/blob/master/CNN/Layer/TransposedConv.py>
 //! - <https://ieee.nitk.ac.in/blog/deconv/>
 use crate::{
-    convolutions::{get_padding_size, im2col_ref, ConvolutionLayer},
+    convolutions::{add_bias, get_padding_size, im2col_ref, ConvolutionLayer},
     ConvKernel, DataRepresentation, Padding,
 };
 use ndarray::*;
@@ -16,17 +16,18 @@ pub struct TransposedConvolutionLayer<F: Float> {
     convolution_layer: ConvolutionLayer<F>,
 }
 
-impl<F: 'static + Float> TransposedConvolutionLayer<F> {
+impl<F: 'static + Float + std::ops::AddAssign> TransposedConvolutionLayer<F> {
     /// Creates new transposed_convolutionLayer. The weights are given in
     /// Pytorch layout.
     /// (in channels, out channels, kernel_height, kernel_width)
     pub fn new(
         weights: ConvKernel<F>,
+        bias: Option<Array1<F>>,
         stride: usize,
         padding: Padding,
     ) -> TransposedConvolutionLayer<F> {
         TransposedConvolutionLayer {
-            convolution_layer: ConvolutionLayer::new(weights, stride, padding),
+            convolution_layer: ConvolutionLayer::new(weights, bias, stride, padding),
         }
     }
 
@@ -35,11 +36,12 @@ impl<F: 'static + Float> TransposedConvolutionLayer<F> {
     /// (kernel height, kernel width, out channels, in channels)
     pub fn new_tf(
         weights: ConvKernel<F>,
+        bias: Option<Array1<F>>,
         stride: usize,
         padding: Padding,
     ) -> TransposedConvolutionLayer<F> {
         TransposedConvolutionLayer {
-            convolution_layer: ConvolutionLayer::new_tf(weights, stride, padding),
+            convolution_layer: ConvolutionLayer::new_tf(weights, bias, stride, padding),
         }
     }
 
@@ -47,6 +49,7 @@ impl<F: 'static + Float> TransposedConvolutionLayer<F> {
     pub fn transposed_convolve(&self, image: &DataRepresentation<F>) -> DataRepresentation<F> {
         let output = conv_transpose2d(
             &self.convolution_layer.kernel,
+            self.convolution_layer.bias.as_ref(),
             &image.view(),
             self.convolution_layer.padding,
             self.convolution_layer.stride,
@@ -73,8 +76,9 @@ impl<F: 'static + Float> TransposedConvolutionLayer<F> {
 /// Returns:
 /// -----------------------------------------------
 /// - out: Output data, of shape (F, H', W')
-pub fn conv_transpose2d<'a, T, V, F: 'static + Float>(
+pub fn conv_transpose2d<'a, T, V, F: 'static + Float + std::ops::AddAssign>(
     kernel_weights: T,
+    bias: Option<&Array1<F>>,
     im2d: V,
     padding: Padding,
     stride: usize,
@@ -197,7 +201,7 @@ where
     let filter_transpose = filter_col_flatten.t();
     let mul = im_col.dot(&filter_transpose); // + bias_m
 
-    if padding == Padding::Same {
+    let output = if padding == Padding::Same {
         let mut mul_reshape = mul
             .into_shape((new_im_height, new_im_width, num_filters))
             .unwrap()
@@ -223,5 +227,6 @@ where
             .into_shape((num_filters, new_im_height, new_im_width))
             .unwrap()
             .into_owned()
-    }
+    };
+    add_bias(&output, bias)
 }
